@@ -10,7 +10,9 @@ import traceback
 
 from azure.storage.queue import QueueClient
 from azure.storage.blob import BlobClient
-from azure.cosmosdb.table import TableService, TableBatch
+from azure.cosmosdb.table import TableService
+from typing import List
+from multiprocessing.pool import ThreadPool
 
 enough_workers_message = "There are already enough active workers"
 workers_long_running = "Workers are active for too long. Going to process as well."
@@ -23,31 +25,34 @@ output_container = "output"
 
 
 class TaskResult:
-    def __init__(self, result, error_message, stack_trace, progress, state):
+    def __init__(self, result, error_message, stack_trace, offset, request_id, state):
         self.result = result
         self.error_message = error_message
         self.stack_trace = stack_trace
-        self.progress = progress
+        self.offset = offset
+        self.request_id = request_id
         self.state = state
 
     @staticmethod
     def create_from_request_entity(entity):
-        result, error, trace, progress, state = None, None, None, None, None
+        result, error, trace, offset, state = None, None, None, None, None
         if 'Result' in entity:
             result = entity['Result']
         if 'Error' in entity:
             error = entity['Error']
         if 'StackTrace' in entity:
             trace = entity['StackTrace']
-        if 'Progress' in entity:
-            progress = entity['Progress']
+        if 'Offset' in entity:
+            offset = entity['Offset']
         if "State" in entity:
             state = entity["State"]
+
+        request_id = "%s_%s" % (entity['PartitionKey'], entity['RowKey'])
         
-        return TaskResult(result, error, trace, progress, state)
+        return TaskResult(result, error, trace, offset, request_id, state)
 
     def to_json(self):
-        json_response = {"Completed": False}
+        json_response = {"Completed": False, "RequestId": self.request_id}
         if self.result:
             json_response['Result'] = json.loads(self.result)
             json_response['Completed'] = True
@@ -55,8 +60,8 @@ class TaskResult:
             json_response['Error'] = self.error_message
             json_response['StackTrace'] = self.stack_trace
             json_response['Completed'] = True
-        if self.progress:
-            json_response['Progress'] = self.progress
+        if self.offset:
+            json_response['Progress'] = self.offset
         if self.state and self.state == "Completed":
             json_response["Completed"] = True
         return json_response
@@ -72,7 +77,7 @@ def singleton(class_):
 
 
 class Worker:
-    def __init__(self, queue_name, metadata_table_name, worker_name, should_wake_up=False, segment_limit=1000, batch_size=100):
+    def __init__(self, queue_name, metadata_table_name, worker_name, should_wake_up=False, segment_limit=200):
         connection_string = os.environ['AzureWebJobsStorage']
         self.metadata_table_name = metadata_table_name
         self.table_client = TableService(connection_string=connection_string)
@@ -86,13 +91,23 @@ class Worker:
         else:
             self.ltb_connection_string = None
         self.segment_limit = segment_limit
-        self.batch_size = batch_size
+        self.pool = ThreadPool(processes=1)
 
     def trigger_blob(self, blob_name, data=None):
         request_id = self._insert_blob_metadata(blob_name, data)
         self._enqueue(request_id)
         self._wake_up()
         return request_id
+    
+    def trigger_blobs(self, blob_names: List[str], data=None):
+        guid = self._generate_id()
+        request_ids = []
+        for blob_name in blob_names:
+            request_id = self._insert_blob_metadata(blob_name, data, guid)
+            self._enqueue(request_id)
+            request_ids.append(request_id)
+        self._wake_up()
+        return request_ids
 
     def trigger(self, data):
         request_id = self._insert_request_metadata(data)
@@ -103,18 +118,12 @@ class Worker:
     def get_result(self, request_id):
         guid, request_hash = self._decode_id(request_id)
         entity = self.table_client.get_entity(self.metadata_table_name, guid, request_hash)
-        if "ChildTasksPartition" in entity and entity['ChildTasksPartition']:
-            child_tasks_partition = entity['ChildTasksPartition']
-            completed_count = 0
-            total_count = 0
-            for task_entity in self.table_client.query_entities(self.metadata_table_name, filter="PartitionKey eq '%s'" % child_tasks_partition):
-                total_count += 1
-                if 'State' in task_entity and task_entity['State'] == 'Completed':
-                    completed_count += 1
-            entity["Progress"] = completed_count * self.segment_limit
-
         return TaskResult.create_from_request_entity(entity)
 
+    def get_results(self, guid):
+        entities = list(self.table_client.query_entities(self.metadata_table_name, filter="PartitionKey eq '%s'" % guid))
+        return [TaskResult.create_from_request_entity(entity) for entity in entities]
+
     def process_messages_stateful(self, func, blob_func=None):
         state_row = list(self.table_client.query_entities(state_table))[0]
         now = datetime.datetime.now(datetime.timezone.utc)
@@ -163,18 +172,13 @@ class Worker:
             else:
                 raise
 
-    # There are 3 scenarious:
-    # 1. Small blob (under segment_limit size) is being processed. Result is written to the output blob. Task state should be "Completed"
-    # 2. Large blob (more than segment_limit) is being processed. Partial result is written to blob_0. Result of blob splitted into segment_list size parts and enqueued.
-    # 3. Parent task (with child_task_partition field). We poll child tasks. Record progress. If child task fails -> parent fails as well.
     def _process_blob(self, request_id, request_entity, func):
-        blob_name, params, child_tasks_partition, retry_count = self._read_metadata(request_entity)
-        if child_tasks_partition:
-            self._poll_child_tasks(request_entity, request_id, child_tasks_partition)
-            return
-
+        blob_name, params, offset, retry_count, batch_count, next_chunk = self._read_metadata(request_entity)
         try:
-            result, whole_blob_read, blobs_iterator = self._process_blob_by_segments(blob_name, func, params)
+            if next_chunk:
+                result, next_chunk = self._process_next_chunk(next_chunk, blob_name, func, offset, params)
+            else:
+                result, next_chunk = self._process_blob_by_segments(blob_name, func, params)
         except Exception as e:
             if retry_count == 3:
                 self._record_failure(request_id, e)
@@ -183,94 +187,31 @@ class Worker:
                 self.table_client.update_entity(self.metadata_table_name, request_entity)
                 self._enqueue(request_id)
             return
-        
+            
         serialized_result = "\n".join(json.dumps(r) for r in result)
-        if whole_blob_read:
-            parts = blob_name.split("/")
-            if parts[0] == self.worker_name:
-                blob_name =  "/".join(parts[1:])
-            out_blob = self._get_output_blob("%s/%s" % (self.worker_name, blob_name))
-
+        if next_chunk:
+            out_blob = self._get_output_blob("%s/%s_%d" % (self.worker_name, blob_name, batch_count))
             out_blob.upload_blob(serialized_result, timeout=1200)
-            request_entity['State'] = 'Completed'
-            self.table_client.update_entity(self.metadata_table_name, request_entity)
-        else:
-            out_blob = self._get_output_blob("%s/%s_0" % (self.worker_name, blob_name))
-            out_blob.upload_blob(serialized_result, timeout=1200)
-            blob_names = self._split_blob_into_chunks(blobs_iterator, blob_name)
-            partition_key = self._enqueue_child_tasks(blob_names, params)
-            request_entity['ChildTasksPartition'] = partition_key
-            request_entity['State'] = 'Polling'
-            self.table_client.update_entity(self.metadata_table_name, request_entity)
-            self._enqueue(request_id)
-
-    def _poll_child_tasks(self, request_entity, request_id, child_tasks_partition):
-        completed_blobs = []
-        total_count = 0
-        for task_entity in self.table_client.query_entities(self.metadata_table_name, filter="PartitionKey eq '%s'" % child_tasks_partition):
-            total_count += 1
-            if 'State' in task_entity and task_entity['State'] == 'Completed':
-                completed_blobs.append(task_entity["BlobName"])
-        if len(completed_blobs) == total_count:
-            self._merge_blobs(completed_blobs, request_entity["BlobName"])
-        else:
-            self._enqueue(request_id)
-
-    def _split_blob_into_chunks(self, segment_iterator, blob_name):
-        blob_names = []
-        batch = []
-        batch_number = 1
-        for source_segment, target_segment in segment_iterator:
-            batch.append([source_segment, target_segment])
-            if len(batch) == self.segment_limit - 1:
-                partial_blob_name = "%s/%s_%d" % (self.worker_name, blob_name, batch_number)
-                out_blob = self._get_input_blob(partial_blob_name)
-                serialized_result = "\n".join([json.dumps(pair) for pair in batch])
-                out_blob.upload_blob(serialized_result, timeout=1200)
-                blob_names.append(partial_blob_name)
-                batch = []
-                batch_number += 1
-        if len(batch):
-            partial_blob_name = "%s/%s_%d" % (self.worker_name, blob_name, batch_number)
-            out_blob = self._get_input_blob(partial_blob_name)
-            serialized_result = "\n".join([json.dumps(pair) for pair in batch])
-            out_blob.upload_blob(serialized_result, timeout=1200)
-            blob_names.append(partial_blob_name)
-        return blob_names
-
-    def _enqueue_child_tasks(self, blob_names, params):
-        partition_key = str(uuid.uuid4()).replace("-", "")
-        request_hashes = [str(hashlib.md5(data.encode()).hexdigest()) for data in blob_names]
-        request_ids = ["%s_%s" % (partition_key, request_hash) for request_hash in request_hashes]
-        batch = TableBatch()
-        batch_size = 0
-        for partial_blob_name, request_hash in zip(blob_names, request_hashes):
-            metadata_entity = {"PartitionKey": partition_key, "RowKey": request_hash, "BlobName": partial_blob_name }
-            if params:
-                metadata_entity['Parameters'] = json.dumps(params)
-            batch.insert_entity(metadata_entity)
-            batch_size += 1
-            if batch_size == 100:
-                self.table_client.commit_batch(self.metadata_table_name, batch)
-                batch_size = 0
-                batch = TableBatch()
                 
-        if batch_size > 0:
-            self.table_client.commit_batch(self.metadata_table_name, batch)
-
-        for child_request_id in request_ids:
-            self._enqueue(child_request_id)
-        return partition_key
-
-    def _merge_blobs(self, blob_names, blob_name):
-        out_blob = self._get_output_blob("%s/%s" % (self.worker_name, blob_name))
-        blob_names.append("%s/%s_0" % (self.worker_name, blob_name))
-        content = []
-        for partial_blob_name in sorted(blob_names, key=lambda b: int(b.split('_')[-1])):
-            in_blob = self._get_output_blob(partial_blob_name)
-            blob_content = in_blob.download_blob().readall().decode()
-            content.append(blob_content)
-        out_blob.upload_blob("\n".join(content), timeout=1200)
+            request_entity['Offset'] = offset + len(result)
+            request_entity['Batch'] = batch_count + 1
+            request_entity['NextChunk'] = json.dumps(next_chunk)
+            self.table_client.update_entity(self.metadata_table_name, request_entity)
+            self._enqueue(request_id)
+        else:
+            out_blob = self._get_output_blob("%s/%s" % (self.worker_name, blob_name))
+            if offset == 0:
+                out_blob.upload_blob(serialized_result, timeout=1200)
+            else:     
+                content = []
+                for batch in range(batch_count):
+                    in_blob = self._get_output_blob("%s/%s_%d" % (self.worker_name, blob_name, batch))
+                    blob_content = in_blob.download_blob().readall().decode()
+                    content.append(blob_content)
+                content.append(serialized_result)
+                out_blob.upload_blob("\n".join(content), timeout=1200)
+            request_entity['State'] = "Completed"
+            self.table_client.update_entity(self.metadata_table_name, request_entity)
 
     def _get_input_blob(self, blob_name):
         return BlobClient.from_connection_string(self.ltb_connection_string, container_name=input_container, blob_name=blob_name)
@@ -280,38 +221,61 @@ class Worker:
 
     def _read_metadata(self, request_entity):
         blob_name = request_entity['BlobName']
-        params = None
+        params, offset, retry_count, batch_count, next_chunk = None, 0, 0, 0, None
         if 'Parameters' in request_entity:
             params = json.loads(request_entity['Parameters'])
-        child_tasks_partition = None
-        if 'ChildTasksPartition' in request_entity:
-            child_tasks_partition = request_entity['ChildTasksPartition']
-        retry_count = 0
+        if 'Offset' in request_entity:
+            offset = request_entity['Offset']
         if 'RetryCount' in request_entity:
             retry_count = request_entity['RetryCount']
-        return blob_name, params, child_tasks_partition, retry_count
+        if 'Batch' in request_entity:
+            batch_count = request_entity['Batch']
+        if 'NextChunk' in request_entity and len(request_entity['NextChunk']) > 0:
+            next_chunk = json.loads(request_entity['NextChunk'])
+        return blob_name, params, offset, retry_count, batch_count, next_chunk
 
     def _process_blob_by_segments(self, blob_name, func, params):
-        result = []
+        result, batch = [], []
         in_blob = self._get_input_blob(blob_name)
-        batch = []
-        whole_blob_read = True
-        segments_processed = 0
         blobs_iterator = self._read_blob_by_segments(in_blob)
-        for source_segment, target_segment in blobs_iterator:
+        for source_segment in blobs_iterator:
             batch.append(source_segment)
-            if len(batch) == self.batch_size:
+            if len(batch) == self.segment_limit:
                 result.extend(func(batch, params))
-                segments_processed += len(batch)
                 batch = []
-                if segments_processed >= self.segment_limit:
-                    whole_blob_read = False
-                    break
+                break
 
-        if len(batch) > 0 and whole_blob_read:
+        if len(batch) > 0: # If segments in blob are less than segment_limit
             result.extend(func(batch, params))
 
-        return result, whole_blob_read, blobs_iterator
+        next_chunk = []
+        for source_segment in blobs_iterator:
+            next_chunk.append(source_segment)
+            if len(next_chunk) == self.segment_limit:
+                break
+
+        return result, next_chunk
+
+    def _process_next_chunk(self, next_chunk, blob_name, func, offset, params):
+        chunk_read_task = self.pool.apply_async(self._get_next_batch, args=(offset + len(next_chunk), blob_name)) # Start reading new chunk
+        result = func(next_chunk, params) # Process current chunk
+        next_chunk = chunk_read_task.get()
+        return result, next_chunk
+
+    def _get_next_batch(self, offset, blob_name):
+        in_blob = self._get_input_blob(blob_name)
+        segment_id = -1
+        next_chunk = []
+        blobs_iterator = self._read_blob_by_segments(in_blob)
+        for source_segment in blobs_iterator:
+            segment_id += 1
+            if segment_id < offset:
+                continue
+
+            next_chunk.append(source_segment)
+            if len(next_chunk) == self.segment_limit:
+                break
+        return next_chunk
 
     def _record_failure(self, request_id, e: Exception):
         track = traceback.format_exc()
@@ -327,13 +291,13 @@ class Worker:
 
     def _insert_request_metadata(self, data):
         serialized_body = json.dumps(data)
-        guid, request_hash, request_id = self._generate_id(serialized_body)
+        guid, request_hash, request_id = self._generate_request_id(serialized_body)
         metadata = {"PartitionKey": guid, "RowKey": request_hash, "Request": serialized_body }
         self.table_client.insert_entity(self.metadata_table_name, metadata)
         return request_id
 
-    def _insert_blob_metadata(self, blob_name, data=None):
-        guid, request_hash, request_id = self._generate_id(blob_name)
+    def _insert_blob_metadata(self, blob_name, data=None, guid=None):
+        guid, request_hash, request_id = self._generate_request_id(blob_name, guid)
         metadata_entity = {"PartitionKey": guid, "RowKey": request_hash, "BlobName": blob_name}
         if data:
             metadata_entity['Parameters'] = json.dumps(data)
@@ -394,25 +358,30 @@ class Worker:
                 l = unicodedata.normalize("NFKD", l).strip()
                 if len(l) == 0:
                     continue
-                source_segment, target_language = json.loads(l)
-                yield source_segment.strip(), target_language.strip()
+                source_segment = json.loads(l)[0].strip()
+                yield source_segment
 
         last_piece = unicodedata.normalize("NFKD", last_piece).strip()
         if len(last_piece) > 0:
             source = None
             try:
-                source, target = json.loads(last_piece)
+                source = json.loads(last_piece)[0].strip()
             except:
                 pass
             if source:
-                yield source.strip(), target.strip()
+                yield source
 
     @staticmethod
-    def _generate_id(data):
-        request_id = str(uuid.uuid4()).replace("-", "")
+    def _generate_request_id(data, guid=None):
+        if not guid:
+            guid = Worker._generate_id()
         request_hash = str(hashlib.md5(data.encode()).hexdigest())
-        response = "%s_%s" % (request_id, request_hash)
-        return request_id, request_hash, response
+        request_id = "%s_%s" % (guid, request_hash)
+        return guid, request_hash, request_id
+
+    @staticmethod
+    def _generate_id():
+        return str(uuid.uuid4()).replace("-", "")
     
     @staticmethod
     def _decode_id(request_id):
